/*++
SandMan framework.
Copyright 2008 (c) Matthieu Suiche. <msuiche[at]gmail.com>

This file is part of SandMan.

SandMan is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

SandMan is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with SandMan.  If not, see <http://www.gnu.org/licenses/>.

Module Name:

    compression.h

Abstract:

    - Windows hibernation compression functions definition.
    - TODO: Windows 2000.

Environment:

    - User mode

Revision History:

    - Matthieu Suiche

--*/
#ifndef _COMPRESSION_
#define _COMPRESSION_

#include <windows.h>
#include <stdio.h>
#include "hiber.h"

#ifdef __cplusplus
extern "C"
{
#endif

//
// xpress encoding!
//
#define XPRESS_MAGIC   "\x81\x81xpress"
#define XPRESS_ENCODE_MAGIC  0x19880922
#define DEFAULT_CONTEXT_SIZE 0x2BD8
#define DELTA_PAGE ((2 * PAGE_SIZE) - 1)
#define UNCOMPRESSED_BLOCK_SIZE (PAGE_SIZE * 0x10)


typedef struct _DECODE_DATA {

    PUCHAR BufferLimit;            // 0x00
    PUCHAR DestinationBuffer1;     // 0x04
    PUCHAR DestinationBuffer2;     // 0x08
    PUCHAR Size2;                  // 0x0C
    ULONG DecodedSize;             // 0x10
    PUCHAR PageLimit;              // 0x14
    PUCHAR DataOffset;             // 0x18
    PUCHAR AdjustedPadding;        // 0x1C
    PUCHAR CompressedBlockSize1;   // 0x20 // PageLimit-1
    PUCHAR CompressedBlockSize2;   // 0x24 // PageLimit-3
    ULONG CompressedBlockSize3;    // 0x28 // PageLimit-1
    PUCHAR PageOffset;             // 0x2C
    ULONG Status;                  // 0x30
    ULONG Status2;                 // 0x34

} DECODE_DATA, *PDECODE_DATA;

typedef struct _HIBER_CONTEXT {

    ULONG Magic;                                  // 0x000
    struct _HIBER_CONTEXT *SelfHiberContext;      // 0x004
    PUCHAR CContextLimit;                         // 0x008
    PUCHAR Buffer2;                               // 0x00C
    PVOID pPopAllocateHiberContextCallback;       // 0x010
    ULONG CompressionType;                        // 0x014
    ULONG u018;                                   // 0x018

} HIBER_CONTEXT, *PHIBER_CONTEXT;

typedef struct _COMPRESS_CONTEXT {

    ULONG u00;               // -0x54
    ULONG CompressionType;   // -0x50
    ULONG u08;               // -0x4c
    PUCHAR u0C;              // -0x48
    PUCHAR u10;              // -0x44
    PUCHAR u14;              // -0x40
    PUCHAR u18;              // -0x3c
    PUCHAR u1C;              // -0x38 Offset Limit
    PULONG u20;              // -0x34
    ULONG Status;            // -0x30 Status
    ULONG u28;               // -0x2c
    ULONG u2C;               // -0x28
    ULONG u30;               // -0x24
    ULONG u34;               // -0x20
    ULONG u38;               // -0x1C
    ULONG u3C;               // -0x18
    PUCHAR Source;           // -0x14
    PUCHAR SourceLimit;      // -0x10
    PUCHAR u48;              // -0x0C
    PUCHAR u4C;              // -0x08
    PUCHAR u50;              // -0x04

} COMPRESS_CONTEXT, *PCOMPRESS_CONTEXT;

//
//
//
__inline static ULONG _bsf(dest, src)
{
ULONG ulRet;
    _asm {
    push ebx
    mov eax, [dest]
    mov ebx, [src]
    bsf eax, ebx
    mov [ulRet], eax  
    pop ebx
    }
    return ulRet;
}

#define _movsb(Dest, Src, SizeToCopy)                   \
{                                                       \
ULONG i;                                                \
                                                        \
    for (i = 0; i < SizeToCopy; i++) Dest[i] = Src[i];  \
                                                        \
    Dest += SizeToCopy;                                 \
    Src += SizeToCopy;                                  \
}

#define _cmpsb(Dest, Src, SizeToCmp, ret)               \
{                                                       \
    ULONG i = 0;                                        \
    while (SizeToCmp > 0)                               \
    {                                                   \
        if (Dest[i] != Src[i]) break;                   \
        i++;                                            \
        SizeToCmp--;                                    \
    }                                                   \
    Dest += i;                                          \
    Src += i;                                           \
    if (SizeToCmp == 0) ret = 1;                        \
}

#define NextXpressBlock(pIXH)                                         \
{                                                                     \
    PtrToUlong(pIXH) += (ULONG)(GetXpressBlockSize( pIXH )            \
                     + (ULONG)sizeof(IMAGE_XPRESS_HEADER));           \
    while (memcmp(pIXH, XPRESS_MAGIC, 8) != 0)                        \
    {                                                                 \
         PtrToUlong(pIXH) += (ULONG)(sizeof(XPRESS_MAGIC) - 1);       \
    }                                                                 \
}

//
// Prototypes.
//
ULONG
XpressDecode(
    ULONG DefaultOffset,
    PUCHAR DestinationBuffer, 
    ULONG Size1, 
    ULONG Size2, 
    PUCHAR BlockEntry,           // @xpress+0x20
    ULONG CompressedBlockSize    // Distance from BlockEntry to NextBlock
);

//
// XpressDecode
//
ULONG
Decode(
    PDECODE_DATA Block
);

//
// XpressEncode
//
ULONG
XpressEncode(
    PHIBER_CONTEXT HiberContext,
    LPVOID CompressionCallback,
    ULONG Size1,
    PUCHAR Source,
    ULONG Size2,
    LPVOID IoResume,
    ULONG Arg7,
    ULONG Arg8
);

//
// XpressEncodeCreate
//
PHIBER_CONTEXT XpressEncodeCreate(
    ULONG Size,
    ULONG Arg2,
    PVOID pPopAllocateHiberContextCallback,
    ULONG CompressionRelated
);

ULONG
Encode(
    PCOMPRESS_CONTEXT CompressContext
);

VOID CompressBuffer(
    LPVOID Callbackfunction,
    LPVOID IoResume,
    ULONG Unused,
    PCOMPRESS_CONTEXT CContext // Additionnal argument
);

VOID
EndCompression(
    PCOMPRESS_CONTEXT UContext
);

ULONG
GetXpressBlockSize(
    PIMAGE_XPRESS_HEADER CompressedBlockHeader
);

#ifdef __cplusplus
}
#endif

#endif